from __future__ import division, absolute_import, print_function

from beets.dbcore import FieldQuery
from beets.plugins import BeetsPlugin


# Copied from beets.dbcore with two additions of 'not'
class NotNoneQuery(FieldQuery):
    """A query that checks whether a field is not null."""

    def col_clause(self):
        return self.field + " IS NOT NULL", ()

    def match(self, item):
        try:
            return item[self.field] is not None
        except KeyError:
            return False

    def __repr__(self):
        return "{0.__class__.__name__}({0.field!r}, {0.fast})".format(self)


class OtherQueriesPlugin(BeetsPlugin):
    def queries(self):
        return {
            '%': NotNoneQuery,
        }



# Copyright 2016, Adrian Sampson.
# Copyright 2019, Christopher Larson.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.

"""Prototyped version of `modify` whose values are defined with a template format string."""

from __future__ import division, absolute_import, print_function

from beets import library, ui
from beets.plugins import BeetsPlugin
from beets.ui import commands, decargs, print_
from beets.ui.commands import _do_query, modify_parse_args, print_and_modify


class ModifyTmplPlugin(BeetsPlugin):
    def __init__(self):
        super(ModifyTmplPlugin, self).__init__()

        self.register_listener('pluginload', self.event_pluginload)

    def event_pluginload(self):
        self.non_album_fields = set(library.Item._media_fields) - set(library.Album.item_keys)

        getters = library.Item._getters()
        getters.update(library.Album._getters())
        getter_fields = list(getters.keys())
        getter_fields.remove('path')
        media_fields = (library.Item._media_fields - library.Item._media_tag_fields)
        self.computed_fields = set(getter_fields) | media_fields

    def commands(self):
        modify_cmd = ui.Subcommand(
            u'modifytmpl', help=u'change metadata fields, evaluating templates', aliases=(u'modt',)
        )
        modify_cmd.parser.add_option(
            u'-m', u'--move', action='store_true', dest='move',
            help=u"move files in the library directory"
        )
        modify_cmd.parser.add_option(
            u'-M', u'--nomove', action='store_false', dest='move',
            help=u"don't move files in library"
        )
        modify_cmd.parser.add_option(
            u'-w', u'--write', action='store_true', default=None,
            help=u"write new metadata to files' tags (default)"
        )
        modify_cmd.parser.add_option(
            u'-W', u'--nowrite', action='store_false', dest='write',
            help=u"don't write metadata (opposite of -w)"
        )
        modify_cmd.parser.add_album_option()
        modify_cmd.parser.add_format_option(target='item')
        modify_cmd.parser.add_option(
            u'-y', u'--yes', action='store_true',
            help=u'skip confirmation'
        )
        modify_cmd.func = self.modify_func
        return [modify_cmd]

    def modify_func(self, lib, opts, args):
        query, mods, dels = modify_parse_args(decargs(args))
        if not mods and not dels:
            raise ui.UserError(u'no modifications specified')
        self.modify_items(lib, mods, dels, query, ui.should_write(opts.write),
                          ui.should_move(opts.move), opts.album, not opts.yes)

    # Copied and tweaked from beets.ui.commands to apply a template
    def modify_items(self, lib, mods, dels, query, write, move, album, confirm):
        """Modifies matching items according to user-specified assignments and
        deletions.

        `mods` is a dictionary of field and value pairse indicating
        assignments. `dels` is a list of fields to be deleted.
        """
        # Parse key=value specifications into a dictionary.
        model_cls = library.Album if album else library.Item

        # Get the items to modify.
        items, albums = _do_query(lib, query, album, False)
        objs = albums if album else items

        reconfirm = self.check_sanity(mods, dels, objs, album)
        if reconfirm:
            confirm = True

        # Apply changes *temporarily*, preview them, and collect modified
        # objects.
        print_(u'Modifying {0} {1}s.'
               .format(len(objs), u'album' if album else u'item'))
        changed = []
        for obj in objs:
            obj_mods = {}
            for key, value in mods.items():
                value = obj.evaluate_template(value)
                obj_mods[key] = model_cls._parse(key, value)

            if print_and_modify(obj, obj_mods, dels) and obj not in changed:
                changed.append(obj)

        # Still something to do?
        if not changed:
            print_(u'No changes to make.')
            return

        # Confirm action.
        if confirm:
            if write and move:
                extra = u', move and write tags'
            elif write:
                extra = u' and write tags'
            elif move:
                extra = u' and move'
            else:
                extra = u''

            changed = ui.input_select_objects(
                u'Really modify%s' % extra, changed,
                lambda o: print_and_modify(o, mods, dels)
            )

        # Apply changes to database and files
        with lib.transaction():
            for obj in changed:
                obj.try_sync(write, move)

    def check_sanity(self, mods, dels, objs, album):
        reconfirm = False
        if not album:
            # TODO: consider changing this from UserError to promotion, automatically applying it to the album
            if any(not obj.singleton for obj in objs):
                for key in mods:
                    if key in library.Album.item_keys:
                        self._log.warning(u'modification of album field `{0}` should be done on the album, not the item', key)
                        reconfirm = True
        else:
            for key in mods:
                if key in self.non_album_fields:
                    self._log.warning(u'modification of non-album field `{0}` should be done on the item, not the album', key)
                    reconfirm = True

        for key in list(mods.keys()) + dels:
            if key in self.computed_fields:
                raise ui.UserError(u'modification or deletion of computed field `{0}` is not supported'.format(key))

        return reconfirm



"""Set 'advisory' and 'albumadvisory' fields on import based on 'itunesadvisory'.

While beets supports extension of mediafile to add new fields, its maintainer
advises against leveraging it, so set flexible fields on import and rely on
those fields independent of the tags.
"""

from __future__ import absolute_import, division, print_function

import mediafile
from beets.dbcore import types
from beets.library import Item
from beets.plugins import BeetsPlugin
from beets.ui import Subcommand, show_model_changes, decargs, print_
from beets.ui.commands import print_and_modify, _do_query


class iTunesAdvisoryPlugin(BeetsPlugin):
    album_types = {
        'albumadvisory': types.INTEGER,
    }
    item_types = {
        'advisory': types.INTEGER,
    }

    def __init__(self):
        super(iTunesAdvisoryPlugin, self).__init__()

        self.config.add({
            'auto': True,
        })

        field = mediafile.MediaField(
            mediafile.MP3DescStorageStyle(u'ITUNESADVISORY'),
            mediafile.MP4StorageStyle(u'rtng'),
            mediafile.StorageStyle(u'ITUNESADVISORY'),
            mediafile.ASFStorageStyle(u'ITUNESADVISORY'),
            out_type=int,
        )
        mediafile.MediaFile.add_field('itunesadvisory', field)

        if self.config['auto'].get():
            self.import_stages = [self.imported]

    def commands(self):
        """Add the commands."""
        write_advisory = Subcommand('write-advisory',
                                    help=u'write the advisory state to tags')
        write_advisory.parser.add_option(
            u'-p', u'--pretend', action='store_true',
            help=u"show all changes but do nothing"
        )
        write_advisory.func = self.write_advisory
        read_advisory = Subcommand('read-advisory',
                                   help=u'read the advisory state from tags')
        read_advisory.parser.add_option(
            u'-p', u'--pretend', action='store_true',
            help=u"show all changes but do nothing"
        )
        read_advisory.func = self.read_advisory
        return [read_advisory, write_advisory]

    def write_advisory(self, lib, opts, args):
        self.config.set_args(opts)
        for item in lib.items(u'advisory:1..'):
            mf = mediafile.MediaFile(item.path)
            if item.advisory != mf.itunesadvisory:
                fakeitem = new_item(item)
                fakeitem.advisory = mf.itunesadvisory
                show_model_changes(item, fakeitem)
                if not opts.pretend:
                    mf.itunesadvisory = item.advisory
                    mf.save()

    def read_advisory(self, lib, opts, args):
        self.config.set_args(opts)
        query = decargs(args)
        items, _ = _do_query(lib, query, None, False)
        if not items:
            print_(u'No items matched the specified query: {0}.'.format(' '.join(query)))
            return
        self.read_items(lib, items, pretend=opts.pretend)

    def imported(self, session, task):
        self.read_items(session.lib, task.imported_items(), task.is_album and task.album or None)

    def read_items(self, lib, items, album=None, pretend=False):
        changed = set()
        albums = {}
        dels = []

        for item in items:
            mf = mediafile.MediaFile(item.path)
            advisory = mf.itunesadvisory
            if advisory:
                if advisory == 1:
                    if album:
                        albums[album.id] = album
                    elif not item.singleton and item.album_id not in albums:
                        albums[item.album_id] = item.get_album()

                if print_and_modify(item, {'advisory': advisory}, dels):
                    changed.add(item)

        for album_id, album in albums.items():
            if print_and_modify(album, {'albumadvisory': 1}, dels):
                changed.add(album)

        if not pretend:
            with lib.transaction():
                for obj in changed:
                    obj.store()


def new_item(item):
    newitem = Item()
    newitem.update(item)
    return newitem



from __future__ import division, absolute_import, print_function

import fnmatch
import os

import beets
from beets.util import bytestring_path, mkdirall
from beetsplug import playlist
from confuse import Filename


class AlternativesPlaylistPlugin(beets.plugins.BeetsPlugin):
    def __init__(self):
        super(AlternativesPlaylistPlugin, self).__init__()
        self.config.add({
            'auto': True,
            'playlist_dir': '.',
            'relative_to': 'library',
            'is_relative': False,
        })

        if self.config['relative_to'].get() not in ['playlist', 'library']:
            self.relative_to = beets.util.bytestring_path(
                self.config['relative_to'].as_filename())
        else:
            self.relative_to = None
        self.is_relative = self.config['is_relative'].get()

        self.alternatives, self.playlist = None, None

        self.register_listener('pluginload', self.pluginload)
        if self.config['auto'].get():
            self.register_listener(
                'alternatives_after_update_with_lib', self.alternatives_after_update)

    def commands(self):
        spl_alt_update = beets.ui.Subcommand(
            'altplaylistupdate',
            help=u'update the playlists for alternatives. Playlist names may be '
            u'passed as arguments. This assumes the alt itself has already been updated.'
        )
        spl_alt_update.func = self.update_cmd
        return [spl_alt_update]

    def update_cmd(self, lib, opts, args):
        if not args:
            raise beets.ui.UserError('must specify alternative')
        args = beets.ui.decargs(args)
        alternative = args.pop(0)
        self.write_playlists(alternative, lib)

    def pluginload(self):
        for plugin in beets.plugins.find_plugins():
            if plugin.name == 'alternatives':
                self.alternatives = plugin
                self.alternatives.update = self.patch_alt_update
            elif plugin.name == 'playlist':
                self.playlist = plugin

        if not self.alternatives:
            raise beets.ui.UserError(
                'alternatives plugin is required for alternativesplaylist')
        if not self.playlist:
            raise beets.ui.UserError(
                'playlist plugin is required for alternativesplaylist')

    def write_playlists(self, alternative, lib):
        for m3u in self.find_playlists():
            try:
                self.update_playlist(lib, alternative, m3u)
            except beets.util.FilesystemError:
                self._log.error('Failed to update playlist: {0}'.format(
                    beets.util.displayable_path(playlist)))

    def alternatives_after_update(self, alternative, lib):
        self.write_playlists(alternative.name, lib)

    def find_playlists(self):
        playlist_dir = beets.util.syspath(self.playlist.playlist_dir)
        for file in find(playlist_dir):
            if fnmatch.fnmatch(file, '*.[mM]3[uU]') or fnmatch.fnmatch(file, '*.[mM]3[uU]8'):
                m3u = beets.util.bytestring_path(file)
                yield m3u

    def update_playlist(self, lib, alternative, m3u):
        playlist_dir = beets.util.bytestring_path(self.playlist.playlist_dir)
        m3uname = os.path.relpath(m3u, playlist_dir)

        outm3u = os.path.join(self.playlist_dir(alternative), m3uname)
        self._log.info('Writing playlist {}'.format(
            beets.util.displayable_path(outm3u)))

        # Gather up the items in the playlist and map to the alternative
        m3ubase, _ = os.path.splitext(m3uname)
        query = playlist.PlaylistQuery(beets.util.as_string(m3ubase))
        pathmap = {}
        for item in lib.items(query):
            pathmap[beets.util.bytestring_path(item.path)] = beets.util.bytestring_path(
                item.get(u'alt.{}'.format(alternative)) or u'')

        src_base_dir = beets.util.bytestring_path(
            self.playlist.relative_to if self.playlist.relative_to
            else os.path.dirname(m3u)
        )

        if not self.relative_to and self.config['relative_to'] == 'library':
            alt_dir = self.alternatives.config[alternative]['directory'].as_filename()
            relative_to = beets.util.bytestring_path(alt_dir)
        else:
            relative_to = self.relative_to

        alt_base_dir = beets.util.bytestring_path(
            relative_to if relative_to
            else os.path.dirname(outm3u)
        )

        lines = []
        with open(m3u, mode='rb') as m3ufile:
            for line in m3ufile:
                srcpath = line.rstrip(b'\r\n')
                is_relative = not os.path.isabs(srcpath)
                if is_relative:
                    srcpath = os.path.join(src_base_dir, srcpath)
                srcpath = beets.util.normpath(srcpath)

                newpath = pathmap.get(srcpath)
                if not newpath:
                    self._log.error(
                        'Failed to map path {} in playlist {} for alt {}', srcpath, m3uname, alternative)
                    return

                if is_relative or self.is_relative:
                    newpath = os.path.relpath(newpath, alt_base_dir)

                lines.append(line.replace(srcpath, newpath))

        mkdirall(outm3u)
        with open(outm3u, 'wb') as m3ufile:
            m3ufile.writelines(lines)

    def playlist_dir(self, alternative):
        alt_dir = self.alternatives.config[alternative]['directory'].as_filename(
        )
        playlist_dir = self.config['playlist_dir'].get(Filename(cwd=alt_dir))
        playlist_dir = bytestring_path(playlist_dir)
        return playlist_dir

    def patch_alt_update(self, lib, options):
        """Tweaked update method for alternatives to send events with the lib object."""
        try:
            alt = self.alternatives.alternative(options.name, lib)
        except KeyError as e:
            raise beets.ui.UserError(u"Alternative collection '{0}' not found."
                                     .format(e.args[0]))
        beets.plugins.send('alternatives_before_update', alternative=alt)
        beets.plugins.send(
            'alternatives_before_update_with_lib', alternative=alt, lib=lib)
        alt.update(create=options.create)
        beets.plugins.send('alternatives_after_update_with_lib',
                           alternative=alt, lib=lib)
        beets.plugins.send('alternatives_after_update', alternative=alt)


def find(dir, dirfilter=None, **walkoptions):
    """ Given a directory, recurse into that directory,
    returning all files as absolute paths. """

    for root, dirs, files in os.walk(dir, **walkoptions):
        if dirfilter is not None:
            for d in dirs:
                if not dirfilter(d):
                    dirs.remove(d)

        for file in files:
            yield os.path.join(root, file)



"""Identify albums whose tracks have inconsistent album fields."""

from __future__ import absolute_import, division, print_function

from beets import ui
from beets.library import Album
from beets.plugins import BeetsPlugin
from beets.ui import decargs, print_


class InconsistentAlbumTracks(BeetsPlugin):
    def __init__(self):
        super(InconsistentAlbumTracks, self).__init__()

        self.config.add({
            'ignored_fields': ['added']
        })

        self.ignored_fields = set(self.config['ignored_fields'].as_str_seq())
        self.album_fields = set(Album.item_keys) - self.ignored_fields

    def commands(self):
        cmd = ui.Subcommand(
            u'inconsistent-album-tracks', help=u'Identify albums whose tracks have inconsistent album fields.'
        )
        cmd.func = self.command
        return [cmd]

    def command(self, lib, opts, args):
        query = decargs(args)
        for album in lib.albums(query):
            for item in album.items():
                for field in self.album_fields:
                    if item[field] != album[field]:
                        print_(u'{}: field `{}` has value `{}` but album value is `{}`'.format(item, field, item[field], album[field]))



"""Record the most recent import via a flexible field `last_import`."""

from __future__ import division, absolute_import, print_function

from beets import config
from beets.plugins import BeetsPlugin
from beets.ui import Subcommand, print_
from beets.dbcore import types


class LastimportPlugin(BeetsPlugin):
    """Record the most recent import via a field."""

    album_types = {
        'last_import': types.BOOLEAN,
    }

    def __init__(self):
        super(LastimportPlugin, self).__init__()

        self.config.add({
            'album': False,
            'format': '',
            'path': False,
        })

        self.import_stages = [self.imported]
        self.register_listener('import_begin', self.clear_last_import)

    def commands(self):
        """Add the last-import command."""
        last_import = Subcommand('last-import',
                                 help=u'print the most recently imported items.')
        last_import.parser.add_all_common_options()
        last_import.func = self.last_import
        return [last_import]

    def last_import(self, lib, opts, args):
        """Print the most recently imported items."""
        self.config.set_args(opts)

        album = self.config['album'].get(bool)
        path = self.config['path'].get(bool)
        fmt = self.config['format'].get(str)

        if path:
            fmt = u'$path'
        elif not fmt:
            if album:
                fmt = u'$added ' + config['format_album'].get()
            else:
                fmt = u'$added ' + config['format_item'].get()

        query = lib.albums if album else lib.items
        for obj in query(u'last_import:1 added+'):
            print_(format(obj, fmt))

    def clear_last_import(self, session):
        """Clear existing last_import fields before the new import."""
        with session.lib.transaction():
            for album in session.lib.albums(u'last_import:1'):
                album['last_import'] = 0
                album.store('last_import')

            for item in session.lib.items(u'last_import:1'):
                item['last_import'] = 0
                item.store('last_import')

    def imported(self, session, task):
        if task.is_album:
            item = task.album
        else:
            item = task.item

        item['last_import'] = 1
        with session.lib.transaction():
            item.store('last_import')



from __future__ import absolute_import, division, print_function

from beets.plugins import BeetsPlugin


class ConvertSingletonsPlugin(BeetsPlugin):
    def __init__(self):
        super(ConvertSingletonsPlugin, self).__init__()

        self.register_listener('import_begin', self.import_begin)
        self.register_listener('album_imported', self.import_album)
        self.register_listener('import', self.import_end)

    def import_begin(self, session):
        self.to_convert = set()

    def import_album(self, lib, album):
        if not album.album or album.album == '[non-album tracks]':
            self.to_convert |= set(album.items())

    def import_end(self, lib, paths):
        albums = {}
        with lib.transaction():
            for item in self.to_convert:
                if item.album_id not in albums:
                    albums[item.album_id] = item.get_album()
                self._log.info('converted {0} to a singleton'.format(item))
                item.album_id = None
                item.store()

        with lib.transaction():
            for album in albums.values():
                if not album.items():
                    self._log.info('removed remnant empty album {0} due to conversions to singleton'.format(album.id))
                    album.remove(with_items=False)



# Do the same calculation as absubmit, but run it only against non-musicbrainz
# tracks, since absubmit/acousticbrainz only handle musicbrainz tracks, and
# store the lowlevel values rather than submitting them.

import confuse
import json
import os
import tempfile
import subprocess

from beets import ui, util, config
from beets.plugins import BeetsPlugin, find_plugins
from beetsplug import acousticbrainz


class ABCalcPlugin(BeetsPlugin):
    supported_extensions = ['.mp3', '.ogg', '.oga', '.flac', '.mp4', '.m4a', '.m4r',
                            '.m4b', '.m4p', '.aac', '.wma', '.asf', '.mpc', '.wv',
                            '.spx', '.tta', '.3g2', '.aif', '.aiff', '.ape']

    def __init__(self):
        super(ABCalcPlugin, self).__init__()

        self.supported_extensions = [util.bytestring_path(ext) for ext in list(self.supported_extensions)]

        self.config.add({
            'force': False,
            'auto': False,
        })

        try:
            self.extractor = config['absubmit']['extractor'].as_str()
        except confuse.ConfigError:
            self.extractor = None

        if self.extractor:
            self.extractor = util.normpath(self.extractor)
            # Expicit path to extractor
            if not os.path.isfile(self.extractor):
                raise ui.UserError(
                    u'Extractor command does not exist: {0}.'.
                    format(self.extractor)
                )
        else:
            # Implicit path to extractor, search for it in path
            self.extractor = 'streaming_extractor_music'

        self.abscheme = dict(acousticbrainz.ABSCHEME)
        del self.abscheme['highlevel']

        try:
            self.tags = config['acousticbrainz']['tags'].as_str_seq()
        except confuse.ConfigError:
            self.tags = None

        if self.config['auto']:
            self.register_listener('import_task_files',
                                   self.import_task_files)

    def commands(self):
        cmd = ui.Subcommand(
            'abcalc',
            help=u'calculate and store lowlevel AcousticBrainz analysis for non-MusicBrainz tracks'
        )
        cmd.parser.add_option(
            u'-f', u'--force', dest='force_recalc',
            action='store_true', default=False,
            help=u're-calculate data when already present'
        )

        def func(lib, opts, args):
            items = lib.items(ui.decargs(args))
            self._get_info(items, ui.should_write(),
                           opts.force_recalc or self.config['force'])

        cmd.func = func
        return [cmd]

    def import_task_files(self, session, task):
        self._get_info(task.imported_items(), False, True)

    def _get_info(self, items, write, force):
        self.absubmit, self.acousticbrainz = None, None

        for plugin in find_plugins():
            if plugin.name == 'absubmit':
                self.absubmit = plugin
            elif plugin.name == 'acousticbrainz':
                self.acousticbrainz = plugin

        if not self.absubmit or not self.acousticbrainz:
            raise ui.UserError('absubmit and acousticbrainz are required for this plugin')

        def func(item):
            return self.analyze(item, write, force)
        util.par_map(func, self.included_items(items, force))

    def included_items(self, items, force):
        for item in items:
            _, ext = os.path.splitext(item.path)
            if ext not in self.supported_extensions:
                self._log.debug(u'skipping item {} as {} files are unsupported', item, ext)
                continue

            if getattr(item, 'mb_trackid', None):
                self._log.debug(u'skipping item {} as it has mb_trackid. use acousticbrainz or absubmit instead', item)
                continue
            elif not force and (not self.tags or 'bpm' in self.tags):
                if getattr(item, 'bpm', 0):
                    self._log.debug(u'skipping item {} as it has bpm. set force to override', item)
                    continue
            yield item

    def get_extractor_data(self, item):
        with tempfile.NamedTemporaryFile() as tmpfile:
            tmpfile.close()

            args = [self.extractor, util.syspath(item.path), tmpfile.name]
            try:
                util.command_output(args)
            except subprocess.CalledProcessError as exc:
                self._log.warning(u'{} "{}" "{}" exited with status {}', self.extractor, util.displayable_path(item.path), tmpfile.name, exc.returncode)
                return

            with open(tmpfile.name, 'rb') as tmp_file:
                return json.load(tmp_file)

    def analyze(self, item, write, force):
        self._log.info(u'getting data for: {0}', item)

        data = self.get_extractor_data(item)
        if not data:
            return

        for attr, val in self.acousticbrainz._map_data_to_scheme(data, self.abscheme):
            if not self.tags or attr in self.tags:
                if not force:
                    if getattr(item, attr, None):
                        self._log.debug(u'attribute {} of {} already set, skipping',
                                        attr,
                                        item)
                        continue

                self._log.debug(u'attribute {} of {} set to {}',
                                attr,
                                item,
                                val)
                setattr(item, attr, val)
            else:
                self._log.debug(u'skipping attribute {} of {}'
                                u' (value {}) due to config',
                                attr,
                                item,
                                val)
        item.store()
        if write:
            item.try_write()



"""Apply modifications on album/item import."""

from __future__ import absolute_import, division, print_function

from beets import ui, util
from beets.library import Album, Item, parse_query_string
from beets.plugins import BeetsPlugin
from beets.ui import decargs
from beets.ui.commands import modify_parse_args, print_and_modify
from beets.util import shlex_split


class ModifyOnImport(BeetsPlugin):
    def __init__(self):
        super(ModifyOnImport, self).__init__()

        self.config.add({
            'modify_album': {},
            'modify_album_items': {},
            'modify_singleton': {},
        })

        self.import_stages = [self.imported]
        self.register_listener('import_begin', self.import_begin)

    def import_begin(self, session):
        self.should_write = ui.should_write()
        self.should_move = ui.should_move()

        for name, model_cls in [('album', Album), ('singleton', Item)]:
            modifies = self.get_modifies(self.config['modify_' + name].items(), model_cls, 'modify_' + name)
            setattr(self, name + '_modifies', modifies)

        self.album_item_modifies = []
        for albumquery, itemmodifies in self.config['modify_album_items'].items():
            albumdbquery, _ = parse_query_string(util.as_string(albumquery), Album)
            modifies = self.get_modifies(itemmodifies.items(), Item, u'modify_album_items.{0}'.format(albumquery))
            self.album_item_modifies.append((albumdbquery, modifies))

    def get_modifies(self, items, model_cls, context):
        modifies = []
        for query, modify in items:
            modify = modify.as_str()
            mod_query, mods, dels = self.parse_modify(modify, model_cls)
            if mod_query:
                raise ui.UserError(u'modifyonimport.{0}["{1}"]: unexpected query `{2}` in value'.format(context, query, mod_query))
            elif not mods and not dels:
                raise ui.UserError(u'modifyonimport.{0}["{1}"]: no modifications found'.format(context, query))
            dbquery, _ = parse_query_string(util.as_string(query), model_cls)
            modifies.append((dbquery, mods, dels))
        return modifies

    def parse_modify(self, modify, model_cls):
        modify = util.as_string(modify)
        args = shlex_split(modify)
        query, mods, dels = modify_parse_args(decargs(args))
        return ' '.join(query), mods, dels

    def imported(self, session, task):
        objs = task.imported_items()
        if task.is_album:
            self.modify_objs(session.lib, [task.album], self.album_modifies, task.is_album)
            # Album modifications can alter the items
            for obj in objs:
                obj.load()

            for albumdbquery, modifies in self.album_item_modifies:
                if albumdbquery.match(task.album):
                    self.modify_objs(session.lib, objs, modifies, is_album=False)
        else:
            self.modify_objs(session.lib, objs, self.singleton_modifies, is_album=False)

    def modify_objs(self, lib, objs, modifies, is_album):
        model_cls = Album if is_album else Item
        changed = []
        for obj in objs:
            all_mods, all_dels = {}, {}
            for dbquery, mods, dels in modifies:
                if dbquery.match(obj):
                    all_mods.update(mods)
                    all_dels.update(dels)

            if all_mods or all_dels:
                obj_mods = {}
                for key, value in all_mods.items():
                    value = obj.evaluate_template(value)
                    obj_mods[key] = model_cls._parse(key, value)

                if print_and_modify(obj, obj_mods, all_dels) and obj not in changed:
                    changed.append(obj)

        with lib.transaction():
            for obj in changed:
                obj.store()

        return changed



"""Queries from beets and the beets documentation."""

from __future__ import division, absolute_import, print_function

from beets.dbcore.query import MatchQuery, NoneQuery, TrueQuery
from beets.plugins import BeetsPlugin


# Via beets.dbcore
class AnyQuery(TrueQuery):
    """Match anything/everything."""

    def __init__(self, pattern):
        super(AnyQuery, self).__init__()


# Via beets.dbcore
class NoneFieldQuery(NoneQuery):
    """Match fields which aren't set / are None."""

    def __init__(self, field, _, fast=True):
        super(NoneFieldQuery, self).__init__(field, fast)

    def match(self, item):
        try:
            return item[self.field] is None
        except KeyError:
            return True


class ExistingQueriesPlugin(BeetsPlugin):
    item_queries = {'any': AnyQuery}
    album_queries = {'any': AnyQuery}

    def queries(self):
        return {
            '^': NoneFieldQuery,
            '@': MatchQuery,
        }



"""Inspect candidate changes when importing.

Add a command to inspect candidate changes and has an option to automatically
inspect the changes when the user selects Apply, and will prompt to continue
when timid.
"""

from __future__ import division, absolute_import, print_function

from beets import autotag, config, importer, library, ui
from beets.plugins import BeetsPlugin
from beets.ui.commands import PromptChoice


def new_item(i):
    ni = library.Item()
    ni.update(i)
    return ni


class ImportInspectPlugin(BeetsPlugin):
    def __init__(self):
        super(ImportInspectPlugin, self).__init__()

        self.config.add({
            'on_apply': True,
            # Fields always ignored
            'ignored': '',
            # Fields ignored only when they're new (no previous value)
            'ignored_new': '',
            # Fields ignored only when they're existing (changed from previous value)
            'ignored_existing': '',
        })

        self.album_fields = set(library.Album.item_keys)
        self.nonalbum_fields = library.Item._media_tag_fields.difference(self.album_fields)
        self.all_fields = list(sorted(self.album_fields | self.nonalbum_fields))

        self.ignored = set(self.config['ignored'].as_str_seq())
        self.album_fields = list(sorted(self.album_fields - self.ignored))
        self.nonalbum_fields = list(sorted(self.nonalbum_fields - self.ignored))

        self.ignored_new = set(self.config['ignored_new'].as_str_seq())
        self.ignored_existing = set(self.config['ignored_existing'].as_str_seq())

        self.register_listener('before_choose_candidate',
                               self.before_choose_candidate_listener)

        if self.config['on_apply'].get():
            self.register_listener('import_task_choice',
                                   self.import_task_choice_listener)

    def before_choose_candidate_listener(self, session, task):
        if task.candidates:
            return [PromptChoice('n', 'iNspect changes',
                                      self.importer_inspect_candidate)]

    def importer_inspect_candidate(self, session, task):
        # Prompt the user for a candidate.
        sel = ui.input_options([], numrange=(1, len(task.candidates)))
        match = task.candidates[sel - 1]

        self.show_changes(session.lib, task, match)
        return None

    def import_task_choice_listener(self, session, task):
        if task.apply:
            if config['import']['quiet'].get():
                return

            changes = self.show_changes(session.lib, task)

            # Only prompt when timid
            if 'timid' in self.config:
                timid = self.config['timid'].get(bool)
            else:
                timid = config['import']['timid'].get()

            if not changes or not timid:
                return

            if not ui.input_yn(u'Continue to apply these changes (Y/n)?'):
                task.set_choice(importer.action.SKIP)

    def show_changes(self, lib, task, match=None):
        if match is None:
            match = task.match

        changes = False
        if task.is_album:
            newmapping = {new_item(item): track_info for item, track_info in match.mapping.items()}
            autotag.apply_metadata(match.info, newmapping)

            # olditems[0].get_album() isn't working, create our own to compare
            olditems = list(match.mapping.keys())
            oldvalues = dict((key, olditems[0][key]) for key in self.album_fields)
            oldalbum = library.Album(lib, **oldvalues)

            newitems = list(newmapping.keys())
            values = dict((key, newitems[0][key]) for key in self.album_fields)
            album = library.Album(lib, **values)
            compare_fields = self.get_fields(self.album_fields, oldvalues)

            album_changes = show_model_changes(album, oldalbum, compare_fields)
            if album_changes:
                changes = True

            new_by_info = {track_info: item for item, track_info in newmapping.items()}
            for item, track_info in match.mapping.items():
                newitem = new_by_info[track_info]
                compare_fields = self.get_fields(self.nonalbum_fields, item)
                item_changes = show_model_changes(newitem, item, compare_fields)
                if item_changes:
                    changes = True
        else:
            fakeitem = new_item(task.item)
            autotag.apply_item_metadata(fakeitem, match.info)
            compare_fields = self.get_fields(self.all_fields, task.item)
            changes = show_model_changes(fakeitem, task.item, compare_fields)

        return changes

    def get_fields(self, keys, old):
        compare_fields = []
        for key in sorted(keys):
            if old.get(key):
                ignored = self.ignored_existing
            else:
                ignored = self.ignored_new
            if key not in ignored and key not in self.ignored:
                compare_fields.append(key)
        return compare_fields


def show_model_changes(new, old=None, fields=None, always=False):
    """Given a Model object, print a list of changes from its pristine
    version stored in the database. Return a boolean indicating whether
    any changes were found.

    `old` may be the "original" object to avoid using the pristine
    version from the database. `fields` may be a list of fields to
    restrict the detection to. `always` indicates whether the object is
    always identified, regardless of whether any changes are present.
    """
    old = old or new._db._get(type(new), new.id)

    # Build up lines showing changed fields.
    changes = []
    for field in sorted(old):
        # Subset of the fields. Never show mtime.
        if field == 'mtime' or (fields and field not in fields):
            continue

        # Detect and show difference for this field.
        line = ui._field_diff(field, old, new)
        if line:
            changes.append(u'  {0}: {1}'.format(field, line))

    # New fields.
    for field in sorted(set(new) - set(old)):
        if fields and field not in fields:
            continue

        changes.append(u'  {0}: {1}'.format(
            field,
            ui.colorize('text_highlight', new.formatted()[field])
        ))

    # Print changes.
    if changes or always:
        ui.print_(format(old))
    if changes:
        ui.print_(u'\n'.join(changes))

    return bool(changes)



"""Add a field to indicate whether an item/album has art embedded."""

from __future__ import division, absolute_import, print_function

from beets import art
from beets.plugins import BeetsPlugin
from beets.dbcore import types
from beets.dbcore.query import FieldQuery


class ItemArtQuery(FieldQuery):
    def match(self, item):
        return bool(art.get_art(self._log, item))


class AllItemsArtQuery(FieldQuery):
    def match(self, album):
        return all(art.get_art(self._log, item) for item in album.items())


class HasArtPlugin(BeetsPlugin):
    def __init__(self):
        super(HasArtPlugin, self).__init__()
        self.item_types = {'has_embedded_art': types.BOOLEAN}
        self.album_types = {'all_tracks_have_art': types.BOOLEAN}

        class PluginItemArtQuery(ItemArtQuery):
            pass
        PluginItemArtQuery._log = self._log

        class PluginAllItemsArtQuery(AllItemsArtQuery):
            pass
        PluginAllItemsArtQuery._log = self._log

        self.item_queries = {'has_embedded_art': PluginItemArtQuery}
        self.album_queries = {'all_tracks_have_art': PluginAllItemsArtQuery}



"""Intercept and block write and move operations on items."""

from __future__ import absolute_import, division, print_function

from beets.library import Item
from beets.plugins import BeetsPlugin
from beets.ui import log
from beets.util import MoveOperation, displayable_path


class NoWritePlugin(BeetsPlugin):
    def __init__(self):
        super(NoWritePlugin, self).__init__()

        # Monkeypatch!
        def nowrite(self, path=None, tags=None, id3v23=None):
            log.warning(u'nowrite: ignoring attempt to write {0}'.format(displayable_path(self.path)))

        def nomove(self, operation=MoveOperation.MOVE, basedir=None,
                   with_album=True, store=True):
            log.warning(u'nowrite: ignoring attempt to move {0}'.format(displayable_path(self.path)))

        Item.write = nowrite
        Item.move = nomove



"""Query albums/items whose items/album match a sub-query.

Note that this is not as performant as would be ideal, but gets the job done. Do
not use this unless it's truly necessary. Most album fields are also accessible
on the item, so use those directly when possible. This will become mostly
unnecessary once https://github.com/beetbox/beets/pull/2988 is merged.

Ex.
```
beet ls album_has:albumtotal:1
beet ls -a any_track_has:genre:foo
beet ls -a all_tracks_have:genre:foo
```

FIXME: rework to use a joined fetch with an inner join query based on the
matching sub-query.
"""

from __future__ import absolute_import, division, print_function

from itertools import groupby

from beets import config
from beets.dbcore import Query, MatchQuery, AndQuery, OrQuery
from beets.dbcore.query import NotQuery
from beets.library import Album, Item, parse_query_string
from beets.plugins import BeetsPlugin
from beets.ui import _open_library


class AlbumQuery(Query):
    """Query items whose album matches a sub-query."""
    _cached_album_results = {}

    def __init__(self, query_string):
        self.query_string = query_string
        if self.query_string in self._cached_album_results:
            albums = self.albums = self._cached_album_results[self.query_string]
        else:
            query, _ = parse_query_string(query_string, Album)
            albums = self.lib.albums(query)
            self._cached_album_results[self.query_string] = self.albums = albums
        self.item_query = OrQuery([MatchQuery('album_id', album.id) for album in albums])

    def clause(self):
        return self.item_query.clause()

    def match(self, item):
        return self.item_query.match(item)

    @classmethod
    def clear_cache(cls):
        if cls._cached_album_results:
            cls._cached_album_results = {}


class AnyTrackQuery(Query):
    """Query albums with at least one track matching a sub-query."""
    _cached_results = {}

    def __init__(self, query_string):
        self.query_string = query_string
        if self.query_string in self._cached_results:
            albums = self.albums = self._cached_results[self.query_string]
        else:
            query, _ = parse_query_string(query_string, Item)
            album_only_query = AndQuery([query, NotQuery(MatchQuery(u'album_id', 0))])
            items = self.lib.items(album_only_query)
            albums = set(i.album_id for i in items)
            self._cached_results[self.query_string] = self.albums = albums
        self.album_query = OrQuery([MatchQuery('id', id) for id in albums])

    def clause(self):
        return self.album_query.clause()

    def match(self, item):
        return self.album_query.match(item)

    @classmethod
    def clear_cache(cls):
        if cls._cached_results:
            cls._cached_results = {}


class AllTrackQuery(Query):
    """Query albums with at least one track matching a sub-query."""
    _cached_results = {}

    def __init__(self, query_string):
        self.query_string = query_string
        if self.query_string in self._cached_results:
            albums = self.albums = self._cached_results[self.query_string]
        else:
            query, _ = parse_query_string(query_string, Item)
            album_only_query = AndQuery([query, NotQuery(MatchQuery(u'album_id', 0))])
            items = self.lib.items(album_only_query)

            def item_album(i):
                return i.album_id

            items_by_album = sorted(items, key=item_album)
            grouped_items = groupby(items_by_album, item_album)
            albums = set()
            for album_id, items in grouped_items:
                items = list(items)
                album = items[0].get_album()
                all_items = album.items()
                if len(items) == len(all_items):
                    albums.add(album_id)

            self._cached_results[self.query_string] = self.albums = albums
        self.album_query = OrQuery([MatchQuery('id', id) for id in albums])

    def clause(self):
        return self.album_query.clause()

    def match(self, item):
        return self.album_query.match(item)

    @classmethod
    def clear_cache(cls):
        if cls._cached_results:
            cls._cached_results = {}


class CrossQueryPlugin(BeetsPlugin):
    def __init__(self):
        super(CrossQueryPlugin, self).__init__()

        class RealAlbumQuery(AlbumQuery):
            lib = _open_library(config)
            log = self._log

        self.item_queries = {'album_has': RealAlbumQuery}

        class RealAnyTrackQuery(AnyTrackQuery):
            lib = _open_library(config)
            log = self._log

        class RealAllTrackQuery(AllTrackQuery):
            lib = _open_library(config)
            log = self._log

        self.album_queries = {
            'any_track_has': RealAnyTrackQuery,
            'all_tracks_have': RealAllTrackQuery,
        }

        self.query_cls = [RealAlbumQuery, RealAnyTrackQuery]

        self.register_listener('pluginload', self.event_pluginload)
        self.register_listener('database_change', self.event_database_change)

    def event_pluginload(self):
        for cls in self.query_cls:
            cls.clear_cache()

    def event_database_change(self, lib, model):
        for cls in self.query_cls:
            cls.clear_cache()



"""Clear fields in albums in the database, obeying the 'zero' plugin configuration.

The zero plugin focuses on updating tags, and incidentally can update the database,
but when it does so, it only updates the files and corresponding items, but not the
album associated to those items. This plugin does so.
"""

from __future__ import absolute_import, division, print_function

from beets import config
from beets.library import Item
from beets.plugins import BeetsPlugin, find_plugins


class ZeroAlbumPlugin(BeetsPlugin):
    def __init__(self):
        super(ZeroAlbumPlugin, self).__init__()

        self.config = config['zero']
        self.fields = list(Item._media_fields)

        self.import_stages = [self.imported]
        self.register_listener('pluginload', self.pluginload)

    def pluginload(self):
        self.zero_plugin = None
        if self.config['auto'] and self.config['update_database']:
            for plugin in find_plugins():
                if plugin.name == 'zero':
                    self.zero_plugin = plugin
                    break

    def imported(self, session, task):
        if self.zero_plugin and task.is_album:
            self.zero_plugin.set_fields(task.album, dict(task.album))
            task.album.store()



"""Intercept and block write and move operations on items."""

from __future__ import absolute_import, division, print_function

from beets import ui
from beets.plugins import BeetsPlugin


class SortedModelChangesPlugin(BeetsPlugin):
    def __init__(self):
        super(SortedModelChangesPlugin, self).__init__()
        self.register_listener('loaded', self.loaded)

    def loaded(self):
        # Monkeypatch
        self._log.warning('loaded')
        ui.show_model_changes = show_model_changes


# Copied from beets.ui
def show_model_changes(new, old=None, fields=None, always=False):
    """Given a Model object, print a list of changes from its pristine
    version stored in the database. Return a boolean indicating whether
    any changes were found.

    `old` may be the "original" object to avoid using the pristine
    version from the database. `fields` may be a list of fields to
    restrict the detection to. `always` indicates whether the object is
    always identified, regardless of whether any changes are present.
    """
    old = old or new._db._get(type(new), new.id)

    # Build up lines showing changed fields.
    changes = []
    for field in old:
        # Subset of the fields. Never show mtime.
        if field == 'mtime' or (fields and field not in fields):
            continue

        # Detect and show difference for this field.
        line = ui._field_diff(field, old, new)
        if line:
            changes.append(u'  {0}: {1}'.format(field, line))

    # New fields.
    for field in set(new) - set(old):
        if fields and field not in fields:
            continue

        changes.append(u'  {0}: {1}'.format(
            field,
            ui.colorize('text_highlight', new.formatted()[field])
        ))

    # Print changes.
    if changes or always:
        ui.print_(format(old))
    if changes:
        ui.print_(u'\n'.join(sorted(changes)))

    return bool(changes)



"""Define saved format strings by storing them in fields."""

from __future__ import division, absolute_import, print_function

from beets import config
from beets.plugins import BeetsPlugin


class SavedFormatsPlugin(BeetsPlugin):
    """Define saved format strings by storing them in fields.

    Similar to inline, except we specify a format string, not code.
    """

    def __init__(self):
        super(SavedFormatsPlugin, self).__init__()

        config.add({
            'item_formats': {},
            'album_formats': {},
        })
        self.set_template_fields()

    def set_template_fields(self):
        for field, template in config['item_formats'].items():
            self._log.debug('adding item field {}', field)

            def apply_item(item, field=field, template=template):
                return self.apply_template(field, template.as_str(), item)
            self.template_fields[field] = apply_item

        for field, template in config['album_formats'].items():
            self._log.debug('adding album field {}', field)

            def apply_album(album, field=field, template=template):
                return self.apply_template(field, template.as_str(), album)
            self.album_template_fields[field] = apply_album

    def apply_template(self, field, template, model):
        return model.evaluate_template(template)
